from typing import Callable, Iterator, Optional, Self, SupportsIndex, Any
from types import NoneType
from collections import deque
import numpy as np

import lucid
import lucid.types as types
from lucid.types import (
    _ArrayOrScalar,
    _NumPyArray,
    _MLXArray,
    _Scalar,
    _DeviceType,
    _BuiltinNumeric,
    Numeric,
)

from lucid._tensor.tensor_ops import _TensorOps
from lucid._backend.metal import mx, parse_mlx_indexing, check_metal_availability


_HookType = Callable[["Tensor", _NumPyArray | _MLXArray], None]

_dtype_map = {int: types.Int64, float: types.Float64, complex: types.Complex64}


def _noop() -> None: ...


class Tensor(_TensorOps):
    def __init__(
        self,
        data: _ArrayOrScalar | _MLXArray,
        requires_grad: bool = False,
        keep_grad: bool = False,
        dtype: _BuiltinNumeric | Numeric | None = None,
        device: _DeviceType = "cpu",
    ) -> None:
        self._is_free = False
        self._is_bool_tensor = False

        if dtype is bool:
            self._is_bool_tensor = True
            dtype = None
        if dtype in {int, float, complex}:
            dtype = _dtype_map[dtype]

        dtype: Numeric | None
        assert isinstance(dtype, (Numeric, NoneType)), type(dtype)

        if dtype is not None and dtype.is_bit_free:
            raise TypeError(
                f"Implicit dtypes is not supported for Tensor instantiation."
            )
        if device not in {"cpu", "gpu"}:
            raise lucid.UnknownDeviceError(device)

        self.data: _NumPyArray | _MLXArray
        if not isinstance(data, (_NumPyArray, _MLXArray)):
            self.data = np.array(data, dtype=dtype.cpu if dtype is not None else dtype)

            if isinstance(data, (_Scalar, list, tuple)):
                self._is_free = True

        elif isinstance(data, _NumPyArray):
            self.data = data
            if dtype is not None and data.dtype != dtype.cpu:
                self.data = data.astype(dtype.cpu)

        elif isinstance(data, _MLXArray):
            check_metal_availability()
            device = "gpu"
            self.data = data
            if dtype is not None and data.dtype != dtype.gpu:
                self.data = data.astype(dtype.gpu)

        if device == "gpu" and isinstance(self.data, _NumPyArray):
            check_metal_availability()
            self.data = mx.array(self.data)

        if "bool" in str(self.data.dtype):
            self._is_bool_tensor = True
        else:
            if self._is_bool_tensor:
                self.data = self.data.astype(bool if device == "cpu" else mx.bool_)

        self._op: object | None = None
        self._backward_op: Callable = _noop
        self._prev: list[Tensor] = []
        self._backward_hooks: list[_HookType] = []

        self.grad: Optional[_NumPyArray | _MLXArray] = None

        self.device = device
        self.requires_grad = requires_grad and lucid.grad_enabled()
        self.keep_grad = keep_grad

        self.dtype: Numeric | bool
        if self._is_bool_tensor:
            self.dtype = bool
        else:
            self.dtype = (
                dtype if dtype is not None else types.to_numeric_type(self.data.dtype)
            )

    @property
    def is_leaf(self) -> bool:
        return self.requires_grad and len(self._prev) == 0

    @property
    def is_free(self) -> bool:
        return self._is_free

    @classmethod
    def is_all_free(cls, *tensors: Self) -> bool:
        return all(t.is_free for t in tensors)

    @classmethod
    def copy_data(cls, data: _NumPyArray | _MLXArray) -> _NumPyArray | _MLXArray:
        if isinstance(data, _NumPyArray):
            return data.copy()
        elif isinstance(data, _MLXArray):
            return mx.array(data)
        else:
            raise TypeError(f"Unexpected type: '{type(data).__name__}'")

    @classmethod
    def copy_grad(cls, grad: _NumPyArray | _MLXArray) -> _NumPyArray | _MLXArray:
        return cls.copy_data(data=grad)

    def free(self) -> Self:
        self._is_free = True
        return self

    def eval(self) -> Self:
        if self.is_gpu():
            mx.eval(self.data)
            mx.stop_gradient(self.data)
        return self

    def backward(self, keep_grad: bool = False) -> None:
        if self.grad is None:
            self.grad = (
                np.ones_like(self.data) if self.is_cpu() else mx.ones_like(self.data)
            )
        visited = set()
        topo_order: list[Self] = []
        stack = [self]

        while stack:
            tensor = stack[-1]
            if tensor in visited:
                stack.pop()
                topo_order.append(tensor)
                continue

            visited.add(tensor)
            for parent in tensor._prev:
                if parent not in visited:
                    stack.append(parent)

        for tensor in reversed(topo_order):
            try:
                tensor._backward_op()
            except Exception as e:
                raise lucid.BackwardError(shape=tensor.shape, op=self._op) from e

            for hook in tensor._backward_hooks:
                hook(tensor, tensor.grad)

            if not (tensor.is_leaf or keep_grad or tensor.keep_grad):
                tensor.grad = None

    def register_hook(self, hook: _HookType) -> Callable:
        self._backward_hooks.append(hook)
        return lambda: self._backward_hooks.remove(hook)

    @property
    def shape(self) -> tuple[int, ...]:
        return self.data.shape

    @property
    def ndim(self) -> int:
        return self.data.ndim

    @property
    def size(self) -> int:
        return self.data.size

    @property
    def flops(self) -> int:
        total = 0
        queue = deque([self])
        visited = set()

        while queue:
            cur = queue.popleft()
            visited.add(cur)
            if cur._op is not None:
                total += cur._op.flops

            for next in cur._prev:
                if next not in visited:
                    queue.appendleft(next)

        return total

    def item(self) -> _Scalar | bool:
        if self.size > 1:
            raise ValueError(
                "Tensor must be 0-dimensional(scalar) to pop its item. "
                "Use `tensor.data` to access the data directly.",
            )
        item = self.data[..., 0] if self.ndim > 0 else self.data
        if self.dtype is bool:
            return bool(item)

        if item % 1 == 0:
            return int(item)
        else:
            return float(item)

    def tolist(self) -> list:
        return self.data.tolist()

    def numpy(self) -> _NumPyArray:
        return np.array(self.data)

    def mlx(self) -> _MLXArray:
        return mx.array(self.data)

    def detach(self) -> Self:
        data_copy = Tensor.copy_data(self.data)
        return Tensor(
            data_copy, requires_grad=False, dtype=self.dtype, device=self.device
        )

    def zero(self) -> None:
        if self.is_cpu():
            self.data = np.zeros_like(self.data)
        else:
            self.data = mx.zeros_like(self.data)

    def zero_grad(self) -> None:
        self.grad = None

    def astype(self, dtype: type | Numeric) -> Self:
        new_dtype = dtype
        if isinstance(dtype, Numeric):
            self._is_bool_tensor = False
            if dtype.is_bit_free:
                new_dtype = dtype.auto_parse(self.data.dtype, device=self.device)
            else:
                new_dtype = dtype.parse(device=self.device)

        elif dtype is bool:
            self._is_bool_tensor = True
            if self.is_gpu():
                new_dtype = mx.bool_

        self.data = self.data.astype(new_dtype)
        if self._is_bool_tensor:
            self.dtype = bool
        else:
            self.dtype = types.to_numeric_type(self.data.dtype)

        return self

    def to(self, device: _DeviceType) -> Self:
        if self.device == device:
            return self

        if device == "cpu":
            self.data = np.array(self.data)
            if self.grad is not None:
                self.grad = np.array(self.grad)

        elif device == "gpu":
            check_metal_availability()
            self.data = mx.array(self.data)
            if self.grad is not None:
                self.grad = mx.array(self.grad)

        else:
            raise lucid.UnknownDeviceError(device)

        self.device = device
        return self

    def is_cpu(self) -> bool:
        return self.device == "cpu"

    def is_gpu(self) -> bool:
        return self.device == "gpu"

    def __getitem__(self, idx: SupportsIndex | Self) -> Self:
        new_idx = idx
        if isinstance(idx, Tensor):
            new_idx = idx.data
            if self.is_gpu() and idx.is_free:
                new_idx = mx.array(idx.data)

        if isinstance(idx, tuple):
            new_idx = tuple()
            for id in idx:
                if isinstance(id, Tensor):
                    if self.is_gpu() and id.is_free:
                        id = mx.array(id.data)
                    else:
                        id = id.data
                new_idx += (id,)

        if self.is_gpu():
            new_idx = parse_mlx_indexing(new_idx)
        else:
            if isinstance(idx, Tensor) and idx.is_gpu():
                raise lucid.DeviceMismatchError(to="gpu", from_="cpu")

        sliced_data = self.data[new_idx]
        new_tensor = Tensor(
            sliced_data, self.requires_grad, self.keep_grad, self.dtype, self.device
        )

        def _backward_op() -> None:
            if self.grad is None:
                self.grad = (
                    np.zeros_like(self.data)
                    if self.is_cpu()
                    else mx.zeros_like(self.data)
                )

            new_grad = lucid._match_grad_shape(
                self.data[new_idx], new_tensor.grad, device=self.device
            )
            lucid._set_tensor_grad(self, new_grad, at=new_idx)

        if self.requires_grad and lucid.grad_enabled():
            new_tensor._backward_op = _backward_op
            new_tensor._prev = [self]

        if self.is_free:
            new_tensor.free()

        return new_tensor

    def __setitem__(
        self, idx: SupportsIndex | Self, value: Self | _ArrayOrScalar
    ) -> None:
        if self.requires_grad:
            raise RuntimeError(
                "Cannot perform in-place item setting on a "
                + "Tensor that requires gradients. "
            )

        new_idx = idx
        if isinstance(idx, Tensor):
            new_idx = idx.data
        elif isinstance(idx, tuple):
            new_idx = tuple()
            for id in idx:
                if isinstance(id, Tensor):
                    id = id.data
                new_idx += (id,)

        if self.is_gpu():
            new_idx = parse_mlx_indexing(new_idx)

        data: _ArrayOrScalar
        if isinstance(value, Tensor):
            if self.device != value.device:
                if value.is_free:
                    data = value.to(self.device).data
                else:
                    raise lucid.DeviceMismatchError(to=self.device, from_=value.device)
            data = value.data
        else:
            data = value
        self.data[new_idx] = data

    def __len__(self) -> int:
        if self.ndim == 0:
            return self.size
        else:
            return self.shape[0]

    def __iter__(self) -> Iterator[Self]:
        for i in range(self.shape[0]):
            yield self[i]

    def __repr__(self) -> str:
        return f"Tensor({self.data}, grad={self.grad}, device={self.device})"

    def __str__(self) -> str:
        return str(self.data)

    def __hash__(self) -> int:
        return hash(id(self))

    def __deepcopy__(self, *args: Any) -> Self:
        copied_data = Tensor.copy_data(self.data)

        new_tensor = Tensor(
            copied_data,
            requires_grad=self.requires_grad,
            keep_grad=self.keep_grad,
            dtype=self.dtype,
            device=self.device,
        )

        if self.grad is not None:
            new_tensor.grad = Tensor.copy_grad(self.grad)

        new_tensor._op = self._op
        new_tensor._backward_op = self._backward_op
        new_tensor._prev = self._prev.copy()
        new_tensor._backward_hooks = self._backward_hooks.copy()

        new_tensor._is_free = self._is_free
        new_tensor._is_bool_tensor = self._is_bool_tensor

        return new_tensor

    def __bool__(self) -> bool:
        if self.data.size != 1:
            raise RuntimeError(
                "Boolean value of Tensor with more than one value is ambiguous. "
                "Use `tensor.any()` or `tensor.all()` instead."
            )
        return bool(self.data.item())

    def any(self, axis: int | None = None, keepdims: bool = False) -> bool | Self:
        if self.is_cpu():
            result = np.any(self.data, axis=axis, keepdims=keepdims)
            return bool(result) if axis is None else Tensor(result, device="cpu")
        else:
            mx.eval(self.data)
            result = mx.any(self.data, axis=axis, keepdims=keepdims)
            return bool(result.item()) if axis is None else Tensor(result, device="gpu")

    def all(self, axis=None, keepdims=False) -> bool | Self:
        if self.is_cpu():
            result = np.all(self.data, axis=axis, keepdims=keepdims)
            return bool(result) if axis is None else Tensor(result, device="cpu")
        else:
            mx.eval(self.data)
            result = mx.all(self.data, axis=axis, keepdims=keepdims)
            return bool(result.item()) if axis is None else Tensor(result, device="gpu")



import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor

from .util import SelectiveSearch, apply_deltas, nms


__all__ = ["RCNN"]


class _RegionWarper(nn.Module):
    def __init__(self, output_size: tuple[int, int] = (224, 224)) -> None:
        super().__init__()
        self.output_size = output_size

    def forward(self, images: Tensor, rois: list[Tensor]) -> Tensor:
        device = images.device
        _, C, H_img, W_img = images.shape

        M = sum(r.shape[0] for r in rois)
        if M == 0:
            return lucid.empty(0, C, *self.output_size, device=device)

        boxes = lucid.concatenate(rois, axis=0).to(device)
        img_ids = lucid.concatenate(
            [lucid.full((len(r),), i, device=device) for i, r in enumerate(rois)]
        )

        widths = boxes[:, 2] - boxes[:, 0] + 1
        heights = boxes[:, 3] - boxes[:, 1] + 1
        ctr_x = boxes[:, 0] + 0.5 * widths
        ctr_y = boxes[:, 1] + 0.5 * heights

        theta = lucid.zeros(M, 2, 3, device=device)
        theta[:, 0, 0] = widths / (W_img - 1)
        theta[:, 1, 1] = heights / (H_img - 1)
        theta[:, 0, 2] = (2 * ctr_x / (W_img - 1)) - 1
        theta[:, 1, 2] = (2 * ctr_y / (H_img - 1)) - 1

        grid = F.affine_grid(theta, size=(M, C, *self.output_size), align_corners=False)
        flat_imgs = images[img_ids]

        return F.grid_sample(flat_imgs, grid, align_corners=False)


class _LinearSVM(nn.Module):
    def __init__(self, feat_dim: int, num_classes: int) -> None:
        super().__init__()
        self.linear = nn.Linear(feat_dim, num_classes)

    def forward(self, x: Tensor) -> Tensor:
        return self.linear(x)

    def get_loss(self, scores: Tensor, labels: Tensor, margin: float = 1.0) -> Tensor:
        N = scores.shape[0]
        correct = scores[lucid.arange(N).to(scores.device), labels].unsqueeze(axis=1)

        margins = F.relu(scores - correct + margin)
        margins[lucid.arange(N).to(scores.device), labels] = 0.0

        return margins.sum() / N


class _BBoxRegressor(nn.Module):
    def __init__(self, feat_dim: int, num_classes: int) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.linear = nn.Linear(feat_dim, num_classes * 4)

    def forward(self, x: Tensor) -> Tensor:
        return self.linear(x).reshape(x.shape[0], self.num_classes, 4)


class RCNN(nn.Module):
    def __init__(
        self,
        backbone: nn.Module,
        feat_dim: int,
        num_classes: int,
        *,
        image_means: tuple[float, float, float] = (0.485, 0.456, 0.406),
        pixel_scale: float = 1.0,
        warper_output_size: tuple[int, int] = (224, 224),
        nms_iou_thresh: float = 0.3,
        score_thresh: float = 0.0,
        add_one: bool = True,
    ) -> None:
        super().__init__()
        self.backbone = backbone
        self.ss = SelectiveSearch()
        self.warper = _RegionWarper(warper_output_size)
        self.svm = _LinearSVM(feat_dim, num_classes)
        self.bbox_reg = _BBoxRegressor(feat_dim, num_classes)

        self.image_means: nn.Buffer
        self.register_buffer(
            "image_means", lucid.Tensor(image_means).reshape(1, 3, 1, 1) / pixel_scale
        )

        self.nms_iou_thresh = nms_iou_thresh
        self.score_thresh = score_thresh
        self.add_one = 1.0 if add_one else 0.0

    def forward(
        self,
        images: Tensor,
        rois: list[Tensor] | None = None,
        *,
        return_feats: bool = False,
    ) -> tuple[Tensor, ...]:
        images = images / lucid.max(images).clip(min_value=1.0)
        images = images - self.image_means

        if rois is None:
            rois = [self.ss(img) for img in images]
        crops = self.warper(images, rois)
        feats = self.backbone(crops)

        if isinstance(feats, (tuple, list)):
            feats = feats[-1]
        feats = feats.flatten(axis=1)

        cls_scores = self.svm(feats)
        bbox_deltas = self.bbox_reg(feats)

        if return_feats:
            return cls_scores, bbox_deltas, feats
        return cls_scores, bbox_deltas

    @lucid.no_grad()
    def predict(
        self, images: Tensor, *, max_det_per_img: int = 100
    ) -> list[dict[str, Tensor]]:
        device = images.device
        _, _, H, W = images.shape

        rois = [self.ss(img) for img in images]
        cls_scores, bbox_deltas = self(images, rois=rois)
        probs = F.softmax(cls_scores, axis=1)

        boxes_all = lucid.concatenate(rois).to(device)
        img_indices = lucid.concatenate(
            [lucid.full((len(r),), i, device=device) for i, r in enumerate(rois)]
        )

        num_classes = probs.shape[1]
        results = [{"boxes": [], "scores": [], "labels": []} for _ in images]

        for c in range(1, num_classes):
            cls_probs = probs[:, c]
            keep_mask = cls_probs > self.score_thresh
            if keep_mask.sum().item() == 0:
                continue

            keep_mask = keep_mask.astype(bool)
            cls_boxes = apply_deltas(
                boxes_all[keep_mask], bbox_deltas[keep_mask, c], self.add_one
            )
            cls_scores = cls_probs[keep_mask]
            cls_imgs = img_indices[keep_mask]

            for img_id in cls_imgs.unique():
                m = cls_imgs == img_id
                det_boxes = cls_boxes[m]
                det_scores = cls_scores[m]

                keep = nms(det_boxes, det_scores, self.nms_iou_thresh)
                if keep.size == 0:
                    continue

                res = results[int(img_id.item())]
                res["boxes"].append(det_boxes[keep])
                res["scores"].append(det_scores[keep])
                res["labels"].append(
                    lucid.full((keep.size,), c, dtype=int, device=device)
                )

        for res in results:
            if not res["boxes"]:
                res["boxes"] = lucid.empty(0, 4, device=device)
                res["scores"] = lucid.empty(0, device=device)
                res["labels"] = lucid.empty(0, dtype=int, device=device)
            else:
                res["boxes"] = lucid.concatenate(res["boxes"])
                res["scores"] = lucid.concatenate(res["scores"])
                res["labels"] = lucid.concatenate(res["labels"])

                if res["scores"].size > max_det_per_img:
                    topk = lucid.topk(res["scores"], k=max_det_per_img)[1]
                    res["boxes"] = res["boxes"][topk]
                    res["scores"] = res["scores"][topk]
                    res["labels"] = res["labels"][topk]

        for res in results:
            b = res["boxes"]
            b = b.clip(min_value=0)
            bx = b[:, [0, 2]].clip(max_value=W - 1)
            by = b[:, [1, 3]].clip(max_value=H - 1)

            res["boxes"] = lucid.concatenate(
                [bx[:, :1], by[:, :1], bx[:, 1:], by[:, 1:]], axis=1
            )

        return results



from typing import Callable
from copy import deepcopy

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor


__all__ = [
    "TransformerEncoderLayer",
    "TransformerDecoderLayer",
    "TransformerEncoder",
    "TransformerDecoder",
    "Transformer",
]


@nn.auto_repr(
    "d_model",
    "num_heads",
    "dim_feedforward",
    "dropout",
    "activation",
    "layer_norm_eps",
    "norm_first",
    "bias",
)
class TransformerEncoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        num_heads: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Callable[[Tensor], Tensor] = F.relu,
        layer_norm_eps: float = 1e-5,
        norm_first: bool = False,
        bias: bool = True,
    ) -> None:
        super().__init__()
        self.self_attn = nn.MultiHeadAttention(
            d_model, num_heads, dropout=dropout, bias=bias
        )

        self.linear1 = nn.Linear(d_model, dim_feedforward, bias=bias)
        self.linear2 = nn.Linear(dim_feedforward, d_model, bias=bias)

        self.dropout = nn.Dropout(dropout)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.norm1 = nn.LayerNorm(d_model, eps=layer_norm_eps)
        self.norm2 = nn.LayerNorm(d_model, eps=layer_norm_eps)

        self.activation = activation
        self.norm_first = norm_first

    def _sa_block(
        self,
        x: Tensor,
        src_mask: Tensor | None,
        src_key_padding_mask: Tensor | None,
        is_causal: bool,
    ) -> Tensor:
        attn_output = self.self_attn(x, x, x, src_key_padding_mask, src_mask, is_causal)
        attn_output = self.dropout1(attn_output)

        return attn_output

    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear1(x)
        x = self.activation(x)
        x = self.dropout(x)
        x = self.linear2(x)
        x = self.dropout2(x)

        return x

    def forward(
        self,
        src: Tensor,
        src_mask: Tensor | None = None,
        src_key_padding_mask: Tensor | None = None,
        is_causal: bool = False,
    ) -> Tensor:
        if self.norm_first:
            x = src + self._sa_block(
                self.norm1(src), src_mask, src_key_padding_mask, is_causal
            )
            x += self._ff_block(self.norm2(x))
        else:
            x = self.norm1(
                src + self._sa_block(src, src_mask, src_key_padding_mask, is_causal)
            )
            x = self.norm2(x + self._ff_block(x))

        return x


@nn.auto_repr(
    "d_model",
    "num_heads",
    "dim_feedforward",
    "dropout",
    "activation",
    "layer_norm_eps",
    "norm_first",
    "bias",
)
class TransformerDecoderLayer(nn.Module):
    def __init__(
        self,
        d_model: int,
        num_heads: int,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Callable[[Tensor], Tensor] = F.relu,
        layer_norm_eps: float = 1e-5,
        norm_first: bool = False,
        bias: bool = True,
    ) -> None:
        super().__init__()
        self.self_attn = nn.MultiHeadAttention(
            d_model, num_heads, dropout=dropout, bias=bias
        )
        self.multihead_attn = nn.MultiHeadAttention(
            d_model, num_heads, dropout=dropout, bias=bias
        )

        self.linear1 = nn.Linear(d_model, dim_feedforward, bias=bias)
        self.linear2 = nn.Linear(dim_feedforward, d_model, bias=bias)

        self.dropout = nn.Dropout(dropout)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.dropout3 = nn.Dropout(dropout)

        self.norm1 = nn.LayerNorm(d_model, eps=layer_norm_eps)
        self.norm2 = nn.LayerNorm(d_model, eps=layer_norm_eps)
        self.norm3 = nn.LayerNorm(d_model, eps=layer_norm_eps)

        self.activation = activation
        self.norm_first = norm_first

    def _sa_block(
        self,
        x: Tensor,
        tgt_mask: Tensor | None,
        tgt_key_padding_mask: Tensor | None,
        is_causal: bool,
    ) -> Tensor:
        attn_output = self.self_attn(x, x, x, tgt_key_padding_mask, tgt_mask, is_causal)
        attn_output = self.dropout1(attn_output)

        return attn_output

    def _mha_block(
        self,
        x: Tensor,
        memory: Tensor,
        mem_mask: Tensor | None,
        mem_key_padding_mask: Tensor | None,
        is_causal: bool,
    ) -> Tensor:
        attn_output = self.multihead_attn(
            x, memory, memory, mem_key_padding_mask, mem_mask, is_causal
        )
        attn_output = self.dropout2(attn_output)

        return attn_output

    def _ff_block(self, x: Tensor) -> Tensor:
        x = self.linear1(x)
        x = self.activation(x)
        x = self.dropout(x)
        x = self.linear2(x)
        x = self.dropout3(x)

        return x

    def forward(
        self,
        tgt: Tensor,
        memory: Tensor,
        tgt_mask: Tensor | None = None,
        mem_mask: Tensor | None = None,
        tgt_key_padding_mask: Tensor | None = None,
        mem_key_padding_mask: Tensor | None = None,
        tgt_is_causal: bool = False,
        mem_is_causal: bool = False,
    ) -> Tensor:
        if self.norm_first:
            x = tgt + self._sa_block(
                self.norm1(tgt), tgt_mask, tgt_key_padding_mask, tgt_is_causal
            )
            x += self._mha_block(
                self.norm2(x), memory, mem_mask, mem_key_padding_mask, mem_is_causal
            )
            x += self._ff_block(self.norm3(x))
        else:
            x = self.norm1(
                tgt + self._sa_block(tgt, tgt_mask, tgt_key_padding_mask, tgt_is_causal)
            )
            x = self.norm2(
                x
                + self._mha_block(
                    x, memory, mem_mask, mem_key_padding_mask, mem_is_causal
                )
            )
            x = self.norm3(x + self._ff_block(x))

        return x


class TransformerEncoder(nn.Module):
    def __init__(
        self,
        encoder_layer: TransformerEncoderLayer | nn.Module,
        num_layers: int,
        norm: nn.Module | None = None,
    ) -> None:
        super().__init__()
        self.layers = nn.ModuleList(
            [deepcopy(encoder_layer) for _ in range(num_layers)]
        )
        self.num_layers = num_layers
        self.norm = norm

    def forward(
        self,
        src: Tensor,
        src_mask: Tensor | None = None,
        src_key_padding_mask: Tensor | None = None,
        is_causal: bool = False,
    ) -> Tensor:
        output = src
        for layer in self.layers:
            output = layer(
                output,
                src_mask=src_mask,
                src_key_padding_mask=src_key_padding_mask,
                is_causal=is_causal,
            )
        if self.norm is not None:
            output = self.norm(output)

        return output

    def extra_repr(self) -> str:
        return f"num_layers={self.num_layers}, norm={self.norm is not None}"


class TransformerDecoder(nn.Module):
    def __init__(
        self,
        decoder_layer: TransformerDecoderLayer | nn.Module,
        num_layers: int,
        norm: nn.Module | None = None,
    ) -> None:
        super().__init__()
        self.layers = nn.ModuleList(
            [deepcopy(decoder_layer) for _ in range(num_layers)]
        )
        self.num_layers = num_layers
        self.norm = norm

    def forward(
        self,
        tgt: Tensor,
        memory: Tensor,
        tgt_mask: Tensor | None = None,
        mem_mask: Tensor | None = None,
        tgt_key_padding_mask: Tensor | None = None,
        mem_key_padding_mask: Tensor | None = None,
        tgt_is_causal: bool = False,
        mem_is_causal: bool = False,
    ) -> Tensor:
        output = tgt
        for layer in self.layers:
            output = layer(
                output,
                memory,
                tgt_mask=tgt_mask,
                mem_mask=mem_mask,
                tgt_key_padding_mask=tgt_key_padding_mask,
                mem_key_padding_mask=mem_key_padding_mask,
                tgt_is_causal=tgt_is_causal,
                mem_is_causal=mem_is_causal,
            )
        if self.norm is not None:
            output = self.norm(output)

        return output

    def extra_repr(self) -> str:
        return f"num_layers={self.num_layers}, norm={self.norm is not None}"


@nn.auto_repr(
    "d_model",
    "num_heads",
    "num_encoder_layers",
    "num_decoder_layers",
    "dim_feedforward",
    "dropout",
    "activation",
    "layer_norm_eps",
    "norm_first",
    "bias",
)
class Transformer(nn.Module):
    def __init__(
        self,
        d_model: int = 512,
        num_heads: int = 8,
        num_encoder_layers: int = 6,
        num_decoder_layers: int = 6,
        dim_feedforward: int = 2048,
        dropout: float = 0.1,
        activation: Callable[[Tensor], Tensor] = F.relu,
        layer_norm_eps: float = 1e-5,
        norm_first: bool = False,
        bias: bool = True,
        custom_encoder: nn.Module | None = None,
        custom_decoder: nn.Module | None = None,
    ) -> None:
        super().__init__()
        if custom_encoder is None:
            encoder_layer = TransformerEncoderLayer(
                d_model,
                num_heads,
                dim_feedforward,
                dropout,
                activation,
                layer_norm_eps,
                norm_first,
                bias,
            )
            self.encoder = TransformerEncoder(
                encoder_layer,
                num_encoder_layers,
                norm=nn.LayerNorm(d_model, eps=layer_norm_eps),
            )
        else:
            self.encoder = custom_encoder

        if custom_decoder is None:
            decoder_layer = TransformerDecoderLayer(
                d_model,
                num_heads,
                dim_feedforward,
                dropout,
                activation,
                layer_norm_eps,
                norm_first,
                bias,
            )
            self.decoder = TransformerDecoder(
                decoder_layer,
                num_decoder_layers,
                norm=nn.LayerNorm(d_model, eps=layer_norm_eps),
            )
        else:
            self.decoder = custom_decoder

    def forward(
        self,
        src: Tensor,
        tgt: Tensor,
        src_mask: Tensor | None = None,
        tgt_mask: Tensor | None = None,
        mem_mask: Tensor | None = None,
        src_key_padding_mask: Tensor | None = None,
        tgt_key_padding_mask: Tensor | None = None,
        mem_key_padding_mask: Tensor | None = None,
    ) -> Tensor:
        memory = self.encoder(
            src,
            src_mask=src_mask,
            src_key_padding_mask=src_key_padding_mask,
            is_causal=False,
        )
        output = self.decoder(
            tgt,
            memory,
            tgt_mask=tgt_mask,
            mem_mask=mem_mask,
            tgt_key_padding_mask=tgt_key_padding_mask,
            mem_key_padding_mask=mem_key_padding_mask,
            tgt_is_causal=False,
            mem_is_causal=False,
        )
        return output



from functools import partial
from types import ModuleType
from typing import Literal
import numpy as np
import math

from lucid._tensor import Tensor
from lucid.types import _Scalar

from lucid._backend.core import (
    operation,
    unary_func_op,
    _FuncOpReturnType,
    _GradFuncType,
)
from lucid._backend.metal import mx


class _pow(operation):
    def __init__(self, exp: _Scalar) -> None:
        self.exp = exp
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data**self.exp)
        return self.result, partial(self.__grad__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data**self.exp)
        return self.result, partial(self.__grad__, a=a)

    def __grad__(self, a: Tensor) -> _GradFuncType:
        return (self.exp * a.data ** (self.exp - 1)) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 11 * a.size


class _neg(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(-a.data)
        return self.result, self.__grad__

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(-a.data)
        return self.result, self.__grad__

    def __grad__(self) -> _GradFuncType:
        return -self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return a.size


class exp(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.exp(a.data))
        return self.result, self.__grad__

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.exp(a.data))
        return self.result, self.__grad__

    def __grad__(self) -> _GradFuncType:
        return self.result.data * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 10 * a.size


class log(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.log(a.data))
        return self.result, partial(self.__grad__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.log(a.data))
        return self.result, partial(self.__grad__, a=a)

    def __grad__(self, a: Tensor) -> _GradFuncType:
        return (1 / a.data) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 10 * a.size


class log2(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.log2(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.log2(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        return (1 / (a.data * lib_.log(2))) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 7 * a.size


class sqrt(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.sqrt(a.data))
        return self.result, self.__grad__

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.sqrt(a.data))
        return self.result, self.__grad__

    def __grad__(self) -> _GradFuncType:
        return (0.5 / self.result.data) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 6 * a.size


class sin(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.sin(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.sin(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        return lib_.cos(a.data) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 6 * a.size


class cos(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.cos(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.cos(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        return -lib_.sin(a.data) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 6 * a.size


class tan(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.tan(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.tan(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        return (1 / (lib_.cos(a.data) ** 2)) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 8 * a.size


class arcsin(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.arcsin(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.arcsin(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        return (1 / lib_.sqrt(1 - a.data**2)) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 4 * a.size


class arccos(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.arccos(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.arccos(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        return (-1 / lib_.sqrt(1 - a.data**2)) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 4 * a.size


class arctan(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.arctan(a.data))
        return self.result, partial(self.__grad__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.arctan(a.data))
        return self.result, partial(self.__grad__, a=a)

    def __grad__(self, a: Tensor) -> _GradFuncType:
        return (1 / (1 + a.data**2)) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 3 * a.size


class sinh(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(np.sinh(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(mx.sinh(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        return lib_.cosh(a.data) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 6 * a.size


class cosh(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(np.cosh(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(mx.cosh(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        return lib_.sinh(a.data) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 6 * a.size


class tanh(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(np.tanh(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(mx.tanh(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        return (1 - lib_.tanh(a.data) ** 2) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 8 * a.size


class clip(operation):
    def __init__(self, min_value: float | None, max_value: float | None) -> None:
        super().__init__()
        self.min_value = min_value
        self.max_value = max_value

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.clip(a.data, self.min_value, self.max_value))
        return self.result, partial(self.__grad_cpu__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.clip(a.data, self.min_value, self.max_value))
        return self.result, partial(self.__grad_gpu__, a=a)

    def __grad_cpu__(self, a: Tensor) -> _GradFuncType:
        grad = np.ones_like(a.data)
        if self.min_value is not None:
            grad[a.data < self.min_value] = 0
        if self.max_value is not None:
            grad[a.data > self.max_value] = 0

        return grad * self.result.grad

    def __grad_gpu__(self, a: Tensor) -> _GradFuncType:
        grad = mx.ones_like(a.data)
        if self.min_value is not None:
            grad = mx.where(a.data < self.min_value, 0, grad)
        if self.max_value is not None:
            grad = mx.where(a.data > self.max_value, 0, grad)

        return grad * self.result.grad


class _abs(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.abs(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.abs(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        return lib_.where(a.data >= 0, 1, -1) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return a.size


class sign(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op(has_gradient=False)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.sign(a.data))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.sign(a.data))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0)

    def __flops__(self, a: Tensor) -> int:
        return a.size


class reciprocal(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(1 / a.data)
        return self.result, partial(self.__grad__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(1 / a.data)
        return self.result, partial(self.__grad__, a=a)

    def __grad__(self, a: Tensor) -> _GradFuncType:
        return (-1 / (a.data**2)) * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return a.size


class square(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.square(a.data))
        return self.result, partial(self.__grad__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.square(a.data))
        return self.result, partial(self.__grad__, a=a)

    def __grad__(self, a: Tensor) -> _GradFuncType:
        return 2 * a.data * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return a.size


class cube(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data**3)
        return self.result, partial(self.__grad__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data**3)
        return self.result, partial(self.__grad__, a=a)

    def __grad__(self, a: Tensor) -> _GradFuncType:
        return 3 * a.data**2 * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return 2 * a.size


class _T(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data.T)
        return self.result, self.__grad__

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data.T)
        return self.result, self.__grad__

    def __grad__(self) -> _GradFuncType:
        return self.result.grad


class _mT(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(a.data.mT)
        return self.result, self.__grad_cpu__

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.swapaxes(a.data, -1, -2))
        return self.result, self.__grad_gpu__

    def __grad_cpu__(self) -> _GradFuncType:
        return self.result.grad.mT

    def __grad_gpu__(self) -> _GradFuncType:
        return mx.swapaxes(self.result.grad, -1, -2)


class transpose(operation):
    def __init__(self, axes: list[int] | None, ndim: int) -> None:
        super().__init__()
        self.axes = self._transpose_axes(axes, ndim)

    def _transpose_axes(self, axes: list[int] | None, ndim: int) -> list:
        if axes is None:
            axes = list(reversed(range(ndim)))
        return axes

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.transpose(a.data, self.axes))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.transpose(a.data, self.axes))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.transpose(self.result.grad, lib_.argsort(lib_.array(self.axes)))


class sum(operation):
    def __init__(self, axis: int | tuple[int] | None, keepdims: bool) -> None:
        super().__init__()
        self.axis = axis
        self.keepdims = keepdims

    @unary_func_op()
    def cpu(self, a: Tensor):
        self.result = Tensor(np.sum(a.data, axis=self.axis, keepdims=self.keepdims))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor):
        self.result = Tensor(mx.sum(a.data, axis=self.axis, keepdims=self.keepdims))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def _grad_shape(self, shape: tuple[int]) -> tuple[int]:
        grad_shape = list(shape)
        if not self.keepdims:
            axis_tuple = self.axis if isinstance(self.axis, tuple) else (self.axis,)
            for ax in axis_tuple:
                grad_shape.insert(ax, 1)

        return tuple(grad_shape)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        if self.axis is None:
            grad = lib_.ones_like(a.data) * self.result.grad
        else:
            grad_shpe = self._grad_shape(self.result.grad.shape)
            grad = lib_.reshape(self.result.grad, grad_shpe)

        return grad

    def __flops__(self, a: Tensor) -> int:
        if self.axis is None:
            return a.size - 1
        if isinstance(self.axis, int):
            self.axis = (self.axis,)

        reduced_size = 1
        for ax in self.axis:
            reduced_size *= a.shape[ax]

        output_size = a.size // reduced_size
        return output_size * (reduced_size - 1)


class trace(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.trace(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.trace(a.data))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        grad = lib_.eye(a.data.shape[0], dtype=a.data.dtype)
        return grad * self.result.grad

    def __flops__(self, a: Tensor) -> int:
        return min(a.shape) - 1


class mean(operation):
    def __init__(self, axis: int | tuple[int] | None, keepdims: bool) -> None:
        super().__init__()
        self.axis = axis
        self.keepdims = keepdims

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.mean(a.data, axis=self.axis, keepdims=self.keepdims))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.mean(a.data, axis=self.axis, keepdims=self.keepdims))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        if self.axis is None:
            count = a.data.size
            grad = lib_.ones_like(a.data) * self.result.grad
        else:
            axis_tuple = self.axis if isinstance(self.axis, tuple) else (self.axis,)
            count = lib_.prod(lib_.array([a.data.shape[ax] for ax in axis_tuple]))

            grad_shape = list(self.result.grad.shape)
            if not self.keepdims:
                for ax in sorted(axis_tuple):
                    grad_shape.insert(ax, 1)

            grad = lib_.reshape(self.result.grad, grad_shape)

        return grad / count

    def __flops__(self, a: Tensor) -> int:
        if self.axis is None:
            return a.size
        if isinstance(self.axis, int):
            self.axis = (self.axis,)

        reduced_size = 1
        for ax in self.axis:
            reduced_size *= a.shape[ax]

        output_size = a.size // reduced_size
        return output_size * reduced_size


class var(operation):
    def __init__(self, axis: int | tuple[int] | None, keepdims: bool) -> None:
        super().__init__()
        self.axis = axis
        self.keepdims = keepdims

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.var(a.data, axis=self.axis, keepdims=self.keepdims))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.var(a.data, axis=self.axis, keepdims=self.keepdims))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        if self.axis is None:
            count = a.data.size
        else:
            axis_tuple = self.axis if isinstance(self.axis, tuple) else (self.axis,)
            count = lib_.prod(lib_.array([a.data.shape[ax] for ax in axis_tuple]))

        mean_val = lib_.mean(a.data, axis=self.axis, keepdims=True)
        grad = (2 / count) * (a.data - mean_val) * self.result.grad

        if self.axis is not None and not self.keepdims:
            grad_shape = list(self.result.grad.shape)
            for ax in sorted(axis_tuple):
                grad_shape.insert(ax, 1)
            grad = lib_.reshape(grad, grad_shape)

        return grad

    def __flops__(self, a: Tensor) -> int:
        if self.axis is None:
            reduced_size = a.size
        else:
            if isinstance(self.axis, int):
                self.axis = (self.axis,)
            reduced_size = 1
            for ax in self.axis:
                reduced_size *= a.shape[ax]

        output_size = a.size // reduced_size
        return output_size * (4 * reduced_size)


class _min_or_max(operation):
    def __init__(
        self, mode: Literal["min", "max"], axis: int | tuple[int] | None, keepdims: bool
    ) -> None:
        super().__init__()
        self.mode = mode
        self.axis = axis
        self.keepdims = keepdims

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = (
            Tensor(np.min(a.data, axis=self.axis, keepdims=self.keepdims))
            if self.mode == "min"
            else Tensor(np.max(a.data, axis=self.axis, keepdims=self.keepdims))
        )
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = (
            Tensor(mx.min(a.data, axis=self.axis, keepdims=self.keepdims))
            if self.mode == "min"
            else Tensor(mx.max(a.data, axis=self.axis, keepdims=self.keepdims))
        )
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        grad = self.result.grad
        if not self.keepdims and self.axis is not None:
            if isinstance(self.axis, tuple):
                for ax in sorted(self.axis):
                    grad = lib_.expand_dims(grad, axis=ax)
            else:
                grad = lib_.expand_dims(grad, axis=self.axis)

        if self.keepdims:
            result_expanded = self.result.data
        else:
            if self.axis is None:
                result_expanded = self.result.data.reshape((1,) * a.data.ndim)
            else:
                if isinstance(self.axis, tuple):
                    result_expanded = self.result.data
                    for ax in sorted(self.axis):
                        result_expanded = lib_.expand_dims(result_expanded, axis=ax)
                else:
                    result_expanded = lib_.expand_dims(self.result.data, axis=self.axis)

        mask = a.data == result_expanded
        counts = lib_.sum(mask, axis=self.axis, keepdims=True)
        counts = lib_.where(counts == 0, 1, counts)

        return mask * grad / counts


class swapaxes(operation):
    def __init__(self, axis1: int, axis2: int) -> None:
        super().__init__()
        self.axis1 = axis1
        self.axis2 = axis2

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.swapaxes(a.data, self.axis1, self.axis2))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.swapaxes(a.data, self.axis1, self.axis2))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.swapaxes(self.result.grad, self.axis1, self.axis2)


class round(operation):
    def __init__(self, decimals: int = 0) -> None:
        super().__init__()
        self.decimals = decimals

    @unary_func_op(has_gradient=False)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.round(a.data, decimals=self.decimals))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.round(a.data, decimals=self.decimals))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0)

    def __flops__(self, a: Tensor) -> int:
        return a.size


class floor(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op(has_gradient=False)
    def cpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(np.floor(a.data))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(mx.floor(a.data))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0)

    def __flops__(self, a: Tensor) -> __init__:
        return a.size


class ceil(operation):
    def __init__(self) -> None:
        super().__init__()

    @unary_func_op(has_gradient=False)
    def cpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(np.ceil(a.data))
        return self.result, partial(self.__grad__, lib_=np)

    @unary_func_op(has_gradient=False, device="gpu")
    def gpu(self, a: Tensor) -> Tensor:
        self.result = Tensor(mx.ceil(a.data))
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradFuncType:
        return lib_.array(0.0)

    def __flops__(self, a: Tensor) -> __init__:
        return a.size


class cumprod(operation):
    def __init__(self, axis: int) -> None:
        super().__init__()
        self.axis = axis

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.cumprod(a.data, axis=self.axis))
        return self.result, partial(self.__grad_cpu__, a=a)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.cumprod(a.data, axis=self.axis))
        return self.result, partial(self.__grad_gpu__, a=a)

    def __grad_cpu__(self, a: Tensor) -> _GradFuncType:
        y = self.result.data
        grad = self.result.grad

        u = grad * y
        rev = np.flip(u, axis=self.axis)
        csum = np.cumsum(rev, axis=self.axis)
        rev_csum = np.flip(csum, axis=self.axis)

        return rev_csum / a.data

    def __grad_gpu__(self, a: Tensor) -> _GradFuncType:
        y = self.result.data
        grad = self.result.grad

        u = grad * y
        slices = [slice(None)] * u.ndim
        slices[self.axis] = slice(None, None, -1)

        rev = u[tuple(slices)]
        csum = mx.cumsum(rev, axis=self.axis)
        rev_csum = csum[tuple(slices)]

        return rev_csum / a.data

    def __flops__(self, a: Tensor) -> int:
        return 4 * a.size


class cumsum(operation):
    def __init__(self, axis: int = -1) -> None:
        super().__init__()
        self.axis = axis

    @unary_func_op()
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(np.cumsum(a.data, axis=self.axis))
        return self.result, partial(self.__grad__, a=a, lib_=np)

    @unary_func_op(device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        self.result = Tensor(mx.cumsum(a.data, axis=self.axis))
        return self.result, partial(self.__grad__, a=a, lib_=mx)

    def __grad__(self, a: Tensor, lib_: ModuleType) -> _GradFuncType:
        g = self.result.grad
        rev = tuple(
            slice(None, None, -1) if ax == self.axis else slice(None)
            for ax in range(a.ndim)
        )
        grad_rev = lib_.cumsum(g[rev], axis=self.axis)
        return grad_rev[rev]

    def __flops__(self, a: Tensor) -> int:
        return a.size



from typing import Callable
import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor

from lucid.models.objdet.util import (
    ROIAlign,
    SelectiveSearch,
    apply_deltas,
    nms,
    clip_boxes,
)


__all__ = ["FastRCNN"]


class FastRCNN(nn.Module):
    def __init__(
        self,
        backbone: nn.Module,
        feat_channels: int,
        num_classes: int,
        pool_size: tuple[int, int] = (7, 7),
        hidden_dim: int = 4096,
        bbox_reg_means: tuple[float, ...] = (0.0, 0.0, 0.0, 0.0),
        bbox_reg_stds: tuple[float, ...] = (0.1, 0.1, 0.2, 0.2),
        dropout: float = 0.5,
        proposal_generator: Callable[..., Tensor] | None = None,
    ) -> None:
        super().__init__()
        self.backbone = backbone
        self.roipool = ROIAlign(output_size=pool_size)
        self.proposal_generator = proposal_generator or SelectiveSearch()

        self.fc1 = nn.Linear(feat_channels * pool_size[0] * pool_size[1], hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.cls_score = nn.Linear(hidden_dim, num_classes)
        self.bbox_pred = nn.Linear(hidden_dim, num_classes * 4)

        self.bbox_reg_means = Tensor(bbox_reg_means, dtype=lucid.Float32)
        self.bbox_reg_stds = Tensor(bbox_reg_stds, dtype=lucid.Float32)

    def forward(
        self,
        images: Tensor,
        rois: Tensor | None = None,
        roi_idx: Tensor | None = None,
        *,
        return_feats: bool = False
    ) -> tuple[Tensor, ...]:
        B, _, H, W = images.shape
        if rois is None or roi_idx is None:
            boxes_list, idx_list = [], []
            for i in range(B):
                props = self.proposal_generator(images[i])
                props_f = props.astype(lucid.Float32)

                norm = props_f / lucid.Tensor([W, H, W, H], dtype=lucid.Float32)
                boxes_list.append(norm)
                idx_list.append(lucid.full((norm.shape[0],), i, dtype=lucid.Int32))

            rois = lucid.concatenate(boxes_list, axis=0)
            roi_idx = lucid.concatenate(idx_list, axis=0)

        feats = self.backbone(images)
        pooled = self.roipool(feats, rois, roi_idx)

        N = pooled.shape[0]
        x = pooled.reshape(N, -1)
        x = F.relu(self.fc1(x))
        x = self.dropout1(x)

        x = F.relu(self.fc2(x))
        x = self.dropout2(x)

        cls_logits = self.cls_score(x)
        bbox_deltas = self.bbox_pred(x)

        if return_feats:
            return cls_logits, bbox_deltas, feats
        return cls_logits, bbox_deltas

    @lucid.no_grad()
    def predict(
        self,
        images: Tensor,
        rois: Tensor | None = None,
        roi_idx: Tensor | None = None,
        score_thresh: float = 0.05,
        nms_thresh: float = 0.3,
        top_k: int = 100,
    ) -> list[dict[str, Tensor]]:
        B, _, H, W = images.shape
        if rois is None or roi_idx is None:
            boxes_list, idx_list = [], []
            for i in range(B):
                props = self.proposal_generator(images[i])
                props_f = props.astype(lucid.Float32)
                boxes_list.append(props_f)
                idx_list.append(lucid.full((props_f.shape[0],), i, dtype=lucid.Int32))

            rois_px = lucid.concatenate(boxes_list, axis=0)
            roi_idx = lucid.concatenate(idx_list, axis=0)

            rois_norm = rois_px / lucid.Tensor([W, H, W, H], dtype=lucid.Float32)
            cls_logits, bbox_deltas = self.forward(images, rois_norm, roi_idx)
        else:
            rois_px = rois
            cls_logits, bbox_deltas = self.forward(images, rois, roi_idx)

        scores = F.softmax(cls_logits, axis=1)
        num_classes = scores.shape[1]
        detections: list[dict[str, Tensor]] = []

        for cl in range(1, num_classes):
            cls_scores = scores[:, cl]
            mask = cls_scores > score_thresh
            if lucid.sum(mask) == 0:
                continue

            deltas_cls = bbox_deltas[:, cl * 4 : (cl + 1) * 4]
            boxes_all = apply_deltas(rois_px, deltas_cls)
            boxes = clip_boxes(boxes_all, (H, W))[mask]

            scores_masked = cls_scores[mask]
            keep = nms(boxes, scores_masked, nms_thresh)[:top_k]

            detections.append(
                {
                    "boxes": boxes[keep],
                    "scores": scores_masked[keep],
                    "labels": lucid.full((keep.shape[0],), cl, dtype=lucid.Int32),
                }
            )

        return detections

    def get_loss(
        self,
        cls_logits: Tensor,
        bbox_deltas: Tensor,
        labels: Tensor,
        reg_targets: Tensor,
    ) -> tuple[Tensor, Tensor, Tensor]:
        cls_loss = F.cross_entropy(cls_logits, labels)
        targets_norm = (reg_targets - self.bbox_reg_means) / self.bbox_reg_stds
        fg_mask = labels > 0

        if lucid.sum(fg_mask) > 0:
            labels_fg = labels[fg_mask]
            deltas_fg = bbox_deltas[fg_mask]

            preds: list[Tensor] = []
            for i, c in enumerate(labels_fg):
                start = c * 4
                preds.append(deltas_fg[i, start : start + 4])

            preds = lucid.stack(preds, axis=0)
            targets_fg = targets_norm[fg_mask]
            reg_loss = F.huber_loss(preds, targets_fg)
        else:
            reg_loss = lucid.zeros((), dtype=lucid.Float32)

        total_loss = cls_loss + reg_loss
        return total_loss, cls_loss, reg_loss



import pickle
from pathlib import Path
from collections import OrderedDict
from typing import Literal

from lucid._tensor import Tensor
from lucid.nn import Module
from lucid.types import _NumPyArray


__all__ = ["save", "load"]

_LucidPortable = Tensor | Module | OrderedDict

FORMAT_VERSION: float = 1.0

EXTENSIONS = Literal[".lct", ".lcd", ".safetensors"]


def save(obj: _LucidPortable, path: Path | str, safetensors: bool = False) -> Path:
    if getattr(obj, "device", None) == "gpu":
        raise RuntimeError(
            f"Only CPU objects ({type(obj).__name__}) " "are able to be saved."
        )

    path = Path(path) if isinstance(path, str) else path
    if safetensors:
        path = path.with_suffix(".safetensors")

    if path.suffix == "":
        if isinstance(obj, Tensor):
            path = path.with_suffix(".lct")
        elif isinstance(obj, (Module, OrderedDict)):
            path = (
                path.with_suffix(".safetensors")
                if safetensors
                else path.with_suffix(".lcd")
            )
        else:
            raise TypeError(
                "Cannot infer file extension: "
                "provide full path or use a recognized type "
                "(Tensor, Module, state_dict)."
            )

    suffix: EXTENSIONS = path.suffix
    if suffix == ".lct":
        if not isinstance(obj, Tensor):
            raise TypeError("Expected a Tensor for .lct file.")
        data = {
            "type": "Tensor",
            "format_version": FORMAT_VERSION,
            "content": obj.numpy(),
        }

    elif suffix == ".lcd":
        if isinstance(obj, Module):
            obj = obj.state_dict()
        if not isinstance(obj, OrderedDict):
            raise TypeError("Expected a state_dict (OrderedDict) for .lcd file.")

        data = {"type": "OrderedDict", "format_version": FORMAT_VERSION, "content": obj}

    elif suffix == ".safetensors":
        try:
            from safetensors.numpy import save_file
        except Exception as e:
            raise ImportError(
                "safetensors is required to save .safetensors files. "
                "Install with `pip install safetensors`."
            ) from e

        if isinstance(obj, Module):
            obj = obj.state_dict()
        if not isinstance(obj, OrderedDict):
            raise TypeError(
                "Expected a state_dict (OrderedDict) for .safetensors file."
            )

        save_file(obj, str(path))
        return path.resolve()

    else:
        raise ValueError(f"Unsupported file extension: {suffix}")

    with open(path, "wb") as f:
        pickle.dump(data, f)

    return path.resolve()


def load(path: Path | str) -> _LucidPortable:
    path = Path(path) if isinstance(path, str) else path

    if not path.exists():
        raise FileNotFoundError(f"No such file: {path}")

    suffix: str = path.suffix

    if suffix == ".safetensors":
        try:
            from safetensors.numpy import load_file
        except Exception as e:
            raise ImportError(
                "safetensors is required to load .safetensors files. "
                "Install with `pip install safetensors`."
            ) from e

        content = load_file(str(path))
        return OrderedDict((k, v) for k, v in content.items())

    with open(path, "rb") as f:
        data = pickle.load(f)

    file_type = data.get("type")
    version = data.get("format_version")

    if version != FORMAT_VERSION:
        raise ValueError(f"Incompatible format version: {version} != {FORMAT_VERSION}")

    if file_type == "Tensor":
        array = data["content"]
        return Tensor(array)

    elif file_type == "OrderedDict":
        return data["content"]

    else:
        raise ValueError(f"Unsupported data type in file: {file_type}")



from typing import Self, Sequence

from lucid.types import _Scalar, _ArrayOrScalar, _ShapeLike, _ArrayLikeInt


class _TensorOps:
    def __add__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __radd__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __sub__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __rsub__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __mul__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __rmul__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __truediv__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __rtruediv__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __matmul__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __eq__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __ne__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __gt__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __ge__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __lt__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __le__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __pow__(self, _: _Scalar) -> Self: ...

    def __neg__(self) -> Self: ...

    def __and__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __rand__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __or__(self, other: Self | _ArrayOrScalar) -> Self: ...

    def __ror__(self, other: Self | _ArrayOrScalar) -> Self: ...

    @property
    def T(self) -> Self: ...

    @property
    def mT(self) -> Self: ...

    def dot(self, other: Self | _ArrayOrScalar) -> Self: ...

    def matmul(self, other: Self) -> Self: ...

    def sum(
        self, axis: int | tuple[int] | None = None, keepdims: bool = False
    ) -> Self: ...

    def mean(
        self, axis: int | tuple[int] | None = None, keepdims: bool = False
    ) -> Self: ...

    def var(
        self, axis: int | tuple[int] | None = None, keepdims: bool = False
    ) -> Self: ...

    def clip(
        self, min_value: _Scalar | None = None, max_value: _Scalar | None = None
    ) -> Self: ...

    def reshape(self, *shape: int) -> Self: ...

    def transpose(self, axes: Sequence[int]) -> Self: ...

    def squeeze(self, axis: _ShapeLike | None = None) -> Self: ...

    def unsqueeze(self, axis: _ShapeLike) -> Self: ...

    def ravel(self) -> Self: ...

    def pad(self, pad_width: _ArrayLikeInt) -> Self: ...

    def repeat(self, repeats: int | Sequence[int], axis: int | None = None) -> Self: ...

    def tile(self, reps: int | Sequence[int]) -> Self: ...

    def flatten(self, start_axis: int = 0, end_axis: int = -1) -> Self: ...

    def split(
        self, size_or_sections: int | Sequence[int], axis: int = 0
    ) -> tuple[Self, ...]: ...

    def tril(self, diagonal: int = 0) -> Self: ...

    def triu(self, diagonal: int = 0) -> Self: ...

    def broadcast_to(self, shape: _ShapeLike) -> Self: ...

    def chunk(self, chunks: int, axis: int = 0) -> tuple[Self, ...]: ...

    def swapaxes(self, axis1: int, axis2: int) -> Self: ...

    def masked_fill(self, mask: Self, value: _Scalar) -> Self: ...

    def roll(
        self, shifts: int | tuple[int, ...], axis: int | tuple[int, ...] | None = None
    ) -> Self: ...

    def round(self, decimals: int = 0) -> Self: ...

    def floor(self) -> Self: ...

    def ceil(self) -> Self: ...

    def unbind(self, axis: int = 0) -> tuple[Self, ...]: ...

    def sort(self, axis: int = -1, descending: bool = False) -> tuple[Self, Self]: ...

    def unique(self, sorted: bool = True, axis: int | None = None) -> Self: ...

    def nonzero(self) -> Self: ...

    def diagonal(self, offset: int = 0, axis1: int = 0, axis2: int = 1) -> Self: ...
